@@ -43,18 +43,14 @@
 #include "test_chessboardgenerator.hpp"
 
 #include <functional>
-#include <limits>
-#include <numeric>
 
-using namespace std;
-using namespace cv;
+namespace opencv_test { namespace {
 
 #define _L2_ERR
 
 //#define DEBUG_CHESSBOARD
 
 #ifdef DEBUG_CHESSBOARD
-#include "opencv2/highgui.hpp"
 void show_points( const Mat& gray, const Mat& expected, const vector<Point2f>& actual, bool was_found )
 {
     Mat rgb( gray.size(), CV_8U);
@@ -77,7 +73,7 @@ void show_points( const Mat& gray, const Mat& expected, const vector<Point2f>& a
 #define show_points(...)
 #endif
 
-enum Pattern { CHESSBOARD, CIRCLES_GRID, ASYMMETRIC_CIRCLES_GRID };
+enum Pattern { CHESSBOARD,CHESSBOARD_SB,CIRCLES_GRID, ASYMMETRIC_CIRCLES_GRID};
 
 class CV_ChessboardDetectorTest : public cvtest::BaseTest
 {
@@ -87,6 +83,10 @@ protected:
     void run(int);
     void run_batch(const string& filename);
     bool checkByGenerator();
+    bool checkByGeneratorHighAccuracy();
+
+    // wraps calls based on the given pattern
+    bool findChessboardCornersWrapper(InputArray image, Size patternSize, OutputArray corners,int flags);
 
     Pattern pattern;
     int algorithmFlags;
@@ -103,7 +103,7 @@ double calcError(const vector<Point2f>& v, const Mat& u)
     int count_exp = u.cols * u.rows;
     const Point2f* u_data = u.ptr<Point2f>();
 
-    double err = numeric_limits<double>::max();
+    double err = std::numeric_limits<double>::max();
     for( int k = 0; k < 2; ++k )
     {
         double err1 = 0;
@@ -146,6 +146,9 @@ void CV_ChessboardDetectorTest::run( int /*start_from */)
         return;*/
     switch( pattern )
     {
+        case CHESSBOARD_SB:
+            checkByGeneratorHighAccuracy();      // not supported by CHESSBOARD
+            /* fallthrough */
         case CHESSBOARD:
             checkByGenerator();
             if (ts->get_err_code() != cvtest::TS::OK)
@@ -187,6 +190,7 @@ void CV_ChessboardDetectorTest::run_batch( const string& filename )
     switch( pattern )
     {
         case CHESSBOARD:
+        case CHESSBOARD_SB:
             folder = string(ts->get_data_path()) + "cv/cameracalibration/";
             break;
         case CIRCLES_GRID:
@@ -202,7 +206,7 @@ void CV_ChessboardDetectorTest::run_batch( const string& filename )
 
     if( !fs.isOpened() || board_list.empty() || !board_list.isSeq() || board_list.size() % 2 != 0 )
     {
-        ts->printf( cvtest::TS::LOG, "%s can not be readed or is not valid\n", (folder + filename).c_str() );
+        ts->printf( cvtest::TS::LOG, "%s can not be read or is not valid\n", (folder + filename).c_str() );
         ts->printf( cvtest::TS::LOG, "fs.isOpened=%d, board_list.empty=%d, board_list.isSeq=%d,board_list.size()%2=%d\n",
             fs.isOpened(), (int)board_list.empty(), board_list.isSeq(), board_list.size()%2);
         ts->set_failed_test_info( cvtest::TS::FAIL_MISSING_TEST_DATA );
@@ -242,24 +246,24 @@ void CV_ChessboardDetectorTest::run_batch( const string& filename )
         Size pattern_size = expected.size();
 
         vector<Point2f> v;
-        bool result = false;
+        int flags = 0;
         switch( pattern )
         {
             case CHESSBOARD:
-                result = findChessboardCorners(gray, pattern_size, v, CALIB_CB_ADAPTIVE_THRESH | CALIB_CB_NORMALIZE_IMAGE);
+                flags = CALIB_CB_ADAPTIVE_THRESH | CALIB_CB_NORMALIZE_IMAGE;
                 break;
             case CIRCLES_GRID:
-                result = findCirclesGrid(gray, pattern_size, v);
-                break;
+            case CHESSBOARD_SB:
             case ASYMMETRIC_CIRCLES_GRID:
-                result = findCirclesGrid(gray, pattern_size, v, CALIB_CB_ASYMMETRIC_GRID | algorithmFlags);
-                break;
+            default:
+                flags = 0;
         }
-
-        if( result ^ doesContatinChessboard || v.size() != count_exp )
+        bool result = findChessboardCornersWrapper(gray, pattern_size,v,flags);
+        if(result ^ doesContatinChessboard || (doesContatinChessboard && v.size() != count_exp))
         {
             ts->printf( cvtest::TS::LOG, "chessboard is detected incorrectly in %s\n", img_file.c_str() );
             ts->set_failed_test_info( cvtest::TS::FAIL_INVALID_OUTPUT );
+            show_points( gray, expected, v, result  );
             return;
         }
 
@@ -336,21 +340,21 @@ bool validateData(const ChessBoardGenerator& cbg, const Size& imgSz,
         {
             const Point2f& cur = mat(i, j);
 
-            tmp = norm( cur - mat(i + 1, j + 1) );
+            tmp = cv::norm(cur - mat(i + 1, j + 1)); // TODO cvtest
             if (tmp < minNeibDist)
-                tmp = minNeibDist;
+                minNeibDist = tmp;
 
-            tmp = norm( cur - mat(i - 1, j + 1 ) );
+            tmp = cv::norm(cur - mat(i - 1, j + 1)); // TODO cvtest
             if (tmp < minNeibDist)
-                tmp = minNeibDist;
+                minNeibDist = tmp;
 
-            tmp = norm( cur - mat(i + 1, j - 1) );
+            tmp = cv::norm(cur - mat(i + 1, j - 1)); // TODO cvtest
             if (tmp < minNeibDist)
-                tmp = minNeibDist;
+                minNeibDist = tmp;
 
-            tmp = norm( cur - mat(i - 1, j - 1) );
+            tmp = cv::norm(cur - mat(i - 1, j - 1)); // TODO cvtest
             if (tmp < minNeibDist)
-                tmp = minNeibDist;
+                minNeibDist = tmp;
         }
 
     const double threshold = 0.25;
@@ -359,6 +363,28 @@ bool validateData(const ChessBoardGenerator& cbg, const Size& imgSz,
     return imgsize * threshold < cbsize;
 }
 
+bool CV_ChessboardDetectorTest::findChessboardCornersWrapper(InputArray image, Size patternSize, OutputArray corners,int flags)
+{
+    switch(pattern)
+    {
+    case CHESSBOARD:
+        return findChessboardCorners(image,patternSize,corners,flags);
+    case CHESSBOARD_SB:
+        // check default settings until flags have been specified
+        return findChessboardCornersSB(image,patternSize,corners,0);
+    case ASYMMETRIC_CIRCLES_GRID:
+        flags |= CALIB_CB_ASYMMETRIC_GRID | algorithmFlags;
+        return findCirclesGrid(image, patternSize,corners,flags);
+    case CIRCLES_GRID:
+        flags |= CALIB_CB_SYMMETRIC_GRID;
+        return findCirclesGrid(image, patternSize,corners,flags);
+    default:
+        ts->printf( cvtest::TS::LOG, "Internal Error: unsupported chessboard pattern" );
+        ts->set_failed_test_info( cvtest::TS::FAIL_GENERIC);
+    }
+    return false;
+}
+
 bool CV_ChessboardDetectorTest::checkByGenerator()
 {
     bool res = true;
@@ -383,6 +409,8 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
     int progress = 0;
     for(int i = 0; i < test_num; ++i)
     {
+        SCOPED_TRACE(cv::format("test_num=%d", test_num));
+
         progress = update_progress( progress, i, test_num, 0 );
         ChessBoardGenerator cbg(sizes[i % sizes_num]);
 
@@ -403,7 +431,7 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
 
         vector<Point2f> corners_found;
         int flags = i % 8; // need to check branches for all flags
-        bool found = findChessboardCorners(cb, cbg.cornersSize(), corners_found, flags);
+        bool found = findChessboardCornersWrapper(cb, cbg.cornersSize(), corners_found, flags);
         if (!found)
         {
             ts->printf( cvtest::TS::LOG, "Chess board corners not found\n" );
@@ -413,19 +441,23 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
         }
 
         double err = calcErrorMinError(cbg.cornersSize(), corners_found, corners_generated);
-        if( err > rough_success_error_level )
+        EXPECT_LE(err, rough_success_error_level) << "bad accuracy of corner guesses";
+#if 0
+        if (err >= rough_success_error_level)
         {
-            ts->printf( cvtest::TS::LOG, "bad accuracy of corner guesses" );
-            ts->set_failed_test_info( cvtest::TS::FAIL_BAD_ACCURACY );
-            res = false;
-            return res;
+            imshow("cb", cb);
+            Mat cb_corners = cb.clone();
+            cv::drawChessboardCorners(cb_corners, cbg.cornersSize(), Mat(corners_found), found);
+            imshow("corners", cb_corners);
+            waitKey(0);
         }
+#endif
     }
 
     /* ***** negative ***** */
     {
         vector<Point2f> corners_found;
-        bool found = findChessboardCorners(bg, Size(8, 7), corners_found);
+        bool found = findChessboardCornersWrapper(bg, Size(8, 7), corners_found,0);
         if (found)
             res = false;
 
@@ -434,18 +466,18 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
         vector<Point2f> cg;
         Mat cb = cbg(bg, camMat, distCoeffs, cg);
 
-        found = findChessboardCorners(cb, Size(3, 4), corners_found);
+        found = findChessboardCornersWrapper(cb, Size(3, 4), corners_found,0);
         if (found)
             res = false;
 
-        Point2f c = std::accumulate(cg.begin(), cg.end(), Point2f(), plus<Point2f>()) * (1.f/cg.size());
+        Point2f c = std::accumulate(cg.begin(), cg.end(), Point2f(), std::plus<Point2f>()) * (1.f/cg.size());
 
         Mat_<double> aff(2, 3);
         aff << 1.0, 0.0, -(double)c.x, 0.0, 1.0, 0.0;
         Mat sh;
         warpAffine(cb, sh, aff, cb.size());
 
-        found = findChessboardCorners(sh, cbg.cornersSize(), corners_found);
+        found = findChessboardCornersWrapper(sh, cbg.cornersSize(), corners_found,0);
         if (found)
             res = false;
 
@@ -455,7 +487,7 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
         cnt.push_back(cg[7+0]); cnt.push_back(cg[7+2]);
         cv::drawContours(cb, cnts, -1, Scalar::all(128), FILLED);
 
-        found = findChessboardCorners(cb, cbg.cornersSize(), corners_found);
+        found = findChessboardCornersWrapper(cb, cbg.cornersSize(), corners_found,0);
         if (found)
             res = false;
 
@@ -465,9 +497,151 @@ bool CV_ChessboardDetectorTest::checkByGenerator()
     return res;
 }
 
+// generates artificial checkerboards using warpPerspective which supports
+// subpixel rendering. The transformation is found by transferring corners to
+// the camera image using a virtual plane.
+bool CV_ChessboardDetectorTest::checkByGeneratorHighAccuracy()
+{
+    // draw 2D pattern
+    cv::Size pattern_size(6,5);
+    int cell_size = 80;
+    bool bwhite = true;
+    cv::Mat image = cv::Mat::ones((pattern_size.height+3)*cell_size,(pattern_size.width+3)*cell_size,CV_8UC1)*255;
+    cv::Mat pimage = image(Rect(cell_size,cell_size,(pattern_size.width+1)*cell_size,(pattern_size.height+1)*cell_size));
+    pimage = 0;
+    for(int row=0;row<=pattern_size.height;++row)
+    {
+        int y = int(cell_size*row+0.5F);
+        bool bwhite2 = bwhite;
+        for(int col=0;col<=pattern_size.width;++col)
+        {
+            if(bwhite2)
+            {
+                int x = int(cell_size*col+0.5F);
+                pimage(cv::Rect(x,y,cell_size,cell_size)) = 255;
+            }
+            bwhite2 = !bwhite2;
+
+        }
+        bwhite = !bwhite;
+    }
+
+    // generate 2d points
+    std::vector<Point2f> pts1,pts2,pts1_all,pts2_all;
+    std::vector<Point3f> pts3d;
+    for(int row=0;row<pattern_size.height;++row)
+    {
+        int y = int(cell_size*(row+2));
+        for(int col=0;col<pattern_size.width;++col)
+        {
+            int x = int(cell_size*(col+2));
+            pts1_all.push_back(cv::Point2f(x-0.5F,y-0.5F));
+        }
+    }
+
+    // back project chessboard corners to a virtual plane
+    double fx = 500;
+    double fy = 500;
+    cv::Point2f center(250,250);
+    double fxi = 1.0/fx;
+    double fyi = 1.0/fy;
+    for(auto &&pt : pts1_all)
+    {
+        // calc camera ray
+        cv::Vec3f ray(float((pt.x-center.x)*fxi),float((pt.y-center.y)*fyi),1.0F);
+        ray /= cv::norm(ray);
+
+        // intersect ray with virtual plane
+        cv::Scalar plane(0,0,1,-1);
+        cv::Vec3f n(float(plane(0)),float(plane(1)),float(plane(2)));
+        cv::Point3f p0(0,0,0);
+
+        cv::Point3f l0(0,0,0);    // camera center in world coordinates
+        p0.z = float(-plane(3)/plane(2));
+        double val1 = ray.dot(n);
+        if(val1 == 0)
+        {
+            ts->printf( cvtest::TS::LOG, "Internal Error: ray and plane are parallel" );
+            ts->set_failed_test_info( cvtest::TS::FAIL_GENERIC);
+            return false;
+        }
+        pts3d.push_back(Point3f(ray/val1*cv::Vec3f((p0-l0)).dot(n))+l0);
+    }
+
+    // generate multiple rotations
+    for(int i=15;i<90;i=i+15)
+    {
+        // project 3d points to new camera
+        Vec3f rvec(0.0F,0.05F,float(float(i)/180.0*CV_PI));
+        Vec3f tvec(0,0,0);
+        cv::Mat k = (cv::Mat_<double>(3,3) << fx/2,0,center.x*2, 0,fy/2,center.y, 0,0,1);
+        cv::projectPoints(pts3d,rvec,tvec,k,cv::Mat(),pts2_all);
+
+        // get perspective transform using four correspondences and wrap original image
+        pts1.clear();
+        pts2.clear();
+        pts1.push_back(pts1_all[0]);
+        pts1.push_back(pts1_all[pattern_size.width-1]);
+        pts1.push_back(pts1_all[pattern_size.width*pattern_size.height-1]);
+        pts1.push_back(pts1_all[pattern_size.width*(pattern_size.height-1)]);
+        pts2.push_back(pts2_all[0]);
+        pts2.push_back(pts2_all[pattern_size.width-1]);
+        pts2.push_back(pts2_all[pattern_size.width*pattern_size.height-1]);
+        pts2.push_back(pts2_all[pattern_size.width*(pattern_size.height-1)]);
+        Mat m2 = getPerspectiveTransform(pts1,pts2);
+        Mat out(image.size(),image.type());
+        warpPerspective(image,out,m2,out.size());
+
+        // find checkerboard
+        vector<Point2f> corners_found;
+        bool found = findChessboardCornersWrapper(out,pattern_size,corners_found,0);
+        if (!found)
+        {
+            ts->printf( cvtest::TS::LOG, "Chess board corners not found\n" );
+            ts->set_failed_test_info( cvtest::TS::FAIL_BAD_ACCURACY );
+            return false;
+        }
+        double err = calcErrorMinError(pattern_size,corners_found,pts2_all);
+        if(err > 0.08)
+        {
+            ts->printf( cvtest::TS::LOG, "bad accuracy of corner guesses" );
+            ts->set_failed_test_info( cvtest::TS::FAIL_BAD_ACCURACY );
+            return false;
+        }
+        //cv::cvtColor(out,out,cv::COLOR_GRAY2BGR);
+        //cv::drawChessboardCorners(out,pattern_size,corners_found,true);
+        //cv::imshow("img",out);
+        //cv::waitKey(-1);
+    }
+    return true;
+}
+
 TEST(Calib3d_ChessboardDetector, accuracy) {  CV_ChessboardDetectorTest test( CHESSBOARD ); test.safe_run(); }
+TEST(Calib3d_ChessboardDetector2, accuracy) {  CV_ChessboardDetectorTest test( CHESSBOARD_SB ); test.safe_run(); }
 TEST(Calib3d_CirclesPatternDetector, accuracy) { CV_ChessboardDetectorTest test( CIRCLES_GRID ); test.safe_run(); }
 TEST(Calib3d_AsymmetricCirclesPatternDetector, accuracy) { CV_ChessboardDetectorTest test( ASYMMETRIC_CIRCLES_GRID ); test.safe_run(); }
+#ifdef HAVE_OPENCV_FLANN
 TEST(Calib3d_AsymmetricCirclesPatternDetectorWithClustering, accuracy) { CV_ChessboardDetectorTest test( ASYMMETRIC_CIRCLES_GRID, CALIB_CB_CLUSTERING ); test.safe_run(); }
+#endif
 
+TEST(Calib3d_CirclesPatternDetectorWithClustering, accuracy)
+{
+    cv::String dataDir = string(TS::ptr()->get_data_path()) + "cv/cameracalibration/circles/";
+
+    cv::Mat expected;
+    FileStorage fs(dataDir + "circles_corners15.dat", FileStorage::READ);
+    fs["corners"] >> expected;
+    fs.release();
+
+    cv::Mat image = cv::imread(dataDir + "circles15.png");
+
+    std::vector<Point2f> centers;
+    cv::findCirclesGrid(image, Size(10, 8), centers, CALIB_CB_SYMMETRIC_GRID | CALIB_CB_CLUSTERING);
+    ASSERT_EQ(expected.total(), centers.size());
+
+    double error = calcError(centers, expected);
+    ASSERT_LE(error, precise_success_error_level);
+}
+
+}} // namespace
 /* End of file. */
