@@ -31,6 +31,8 @@
 #ifndef OPENCV_FLANN_BASE_HPP_
 #define OPENCV_FLANN_BASE_HPP_
 
+//! @cond IGNORED
+
 #include <vector>
 #include <cassert>
 #include <cstdio>
@@ -128,7 +130,7 @@ public:
     /**
      * Builds the index.
      */
-    void buildIndex()
+    void buildIndex() CV_OVERRIDE
     {
         if (!loaded_) {
             nnIndex_->buildIndex();
@@ -150,7 +152,7 @@ public:
      * \brief Saves the index to a stream
      * \param stream The stream to save the index to
      */
-    virtual void saveIndex(FILE* stream)
+    virtual void saveIndex(FILE* stream) CV_OVERRIDE
     {
         nnIndex_->saveIndex(stream);
     }
@@ -159,7 +161,7 @@ public:
      * \brief Loads the index from a stream
      * \param stream The stream from which the index is loaded
      */
-    virtual void loadIndex(FILE* stream)
+    virtual void loadIndex(FILE* stream) CV_OVERRIDE
     {
         nnIndex_->loadIndex(stream);
     }
@@ -167,7 +169,7 @@ public:
     /**
      * \returns number of features in this index.
      */
-    size_t veclen() const
+    size_t veclen() const CV_OVERRIDE
     {
         return nnIndex_->veclen();
     }
@@ -175,7 +177,7 @@ public:
     /**
      * \returns The dimensionality of the features in this index.
      */
-    size_t size() const
+    size_t size() const CV_OVERRIDE
     {
         return nnIndex_->size();
     }
@@ -183,7 +185,7 @@ public:
     /**
      * \returns The index type (kdtree, kmeans,...)
      */
-    flann_algorithm_t getType() const
+    flann_algorithm_t getType() const CV_OVERRIDE
     {
         return nnIndex_->getType();
     }
@@ -191,7 +193,7 @@ public:
     /**
      * \returns The amount of memory (in bytes) used by the index.
      */
-    virtual int usedMemory() const
+    virtual int usedMemory() const CV_OVERRIDE
     {
         return nnIndex_->usedMemory();
     }
@@ -200,7 +202,7 @@ public:
     /**
      * \returns The index parameters
      */
-    IndexParams getParameters() const
+    IndexParams getParameters() const CV_OVERRIDE
     {
         return nnIndex_->getParameters();
     }
@@ -213,7 +215,7 @@ public:
      * \param[in] knn Number of nearest neighbors to return
      * \param[in] params Search parameters
      */
-    void knnSearch(const Matrix<ElementType>& queries, Matrix<int>& indices, Matrix<DistanceType>& dists, int knn, const SearchParams& params)
+    void knnSearch(const Matrix<ElementType>& queries, Matrix<int>& indices, Matrix<DistanceType>& dists, int knn, const SearchParams& params) CV_OVERRIDE
     {
         nnIndex_->knnSearch(queries, indices, dists, knn, params);
     }
@@ -227,7 +229,7 @@ public:
      * \param[in] params Search parameters
      * \returns Number of neighbors found
      */
-    int radiusSearch(const Matrix<ElementType>& query, Matrix<int>& indices, Matrix<DistanceType>& dists, float radius, const SearchParams& params)
+    int radiusSearch(const Matrix<ElementType>& query, Matrix<int>& indices, Matrix<DistanceType>& dists, float radius, const SearchParams& params) CV_OVERRIDE
     {
         return nnIndex_->radiusSearch(query, indices, dists, radius, params);
     }
@@ -235,7 +237,7 @@ public:
     /**
      * \brief Method that searches for nearest-neighbours
      */
-    void findNeighbors(ResultSet<DistanceType>& result, const ElementType* vec, const SearchParams& searchParams)
+    void findNeighbors(ResultSet<DistanceType>& result, const ElementType* vec, const SearchParams& searchParams) CV_OVERRIDE
     {
         nnIndex_->findNeighbors(result, vec, searchParams);
     }
@@ -264,6 +266,9 @@ private:
     bool loaded_;
     /** Parameters passed to the index */
     IndexParams index_params_;
+
+    Index(const Index &); // copy disabled
+    Index& operator=(const Index &); // assign disabled
 };
 
 /**
@@ -289,4 +294,7 @@ int hierarchicalClustering(const Matrix<typename Distance::ElementType>& points,
 }
 
 }
+
+//! @endcond
+
 #endif /* OPENCV_FLANN_BASE_HPP_ */
